"""Various implementations of the Lomb-Scargle Periodogram."""

from .chi2_impl import lombscargle_chi2
from .fast_impl import lombscargle_fast
from .fastchi2_impl import lombscargle_fastchi2
from .main import available_methods, lombscargle
from .scipy_impl import lombscargle_scipy
from .slow_impl import lombscargle_slow
