.. This (-*- rst -*-) format file contains commonly used link targets
   and name substitutions.  It may be included in many files,
   therefore it should only contain link targets and name
   substitutions.  Try grepping for "^\.\. _" to find plausible
   candidates for this list.

.. NOTE: reST targets are
   __not_case_sensitive__, so only one target definition is needed for
   nipy, NIPY, Nipy, etc...

.. git stuff
.. _git: https://git-scm.com/
.. _github: https://github.com/
.. _GitHub Help: https://help.github.com/
.. _msysgit: http://code.google.com/p/msysgit/downloads/list
.. _git-osx-installer: http://code.google.com/p/git-osx-installer/downloads/list
.. _git cheat sheet: https://www.atlassian.com/git/tutorials/atlassian-git-cheatsheet
.. _pro git book: https://git-scm.com/book
.. _git svn crash course: https://git.wiki.kernel.org/index.php/GitSvnCrashCourse
.. _learn.github: https://services.github.com/
.. _network graph visualizer: https://github.blog/2008-04-10-say-hello-to-the-network-graph-visualizer/
.. _git user manual: http://schacon.github.io/git/user-manual.html
.. _git tutorial: http://schacon.github.io/git/gittutorial.html
.. _git community book: https://book.git-scm.com/
.. _git casts: https://services.github.com/
.. _git concepts: https://www.sbf5.com/~cduan/technical/git/
.. _git clone: http://schacon.github.io/git/git-clone.html
.. _git checkout: http://schacon.github.io/git/git-checkout.html
.. _git commit: http://schacon.github.io/git/git-commit.html
.. _git push: http://schacon.github.io/git/git-push.html
.. _git pull: http://schacon.github.io/git/git-pull.html
.. _git add: http://schacon.github.io/git/git-add.html
.. _git status: http://schacon.github.io/git/git-status.html
.. _git diff: http://schacon.github.io/git/git-diff.html
.. _git log: http://schacon.github.io/git/git-log.html
.. _git branch: http://schacon.github.io/git/git-branch.html
.. _git remote: http://schacon.github.io/git/git-remote.html
.. _git rebase: http://schacon.github.io/git/git-rebase.html
.. _git config: http://schacon.github.io/git/git-config.html
.. _tangled working copy problem: http://tomayko.com/writings/the-thing-about-git
.. _git management: https://web.archive.org/web/20120511084711/http://kerneltrap.org/Linux/Git_Management
.. _linux git workflow: http://www.mail-archive.com/dri-devel@lists.sourceforge.net/msg39091.html
.. _git parable: https://tom.preston-werner.com/2009/05/19/the-git-parable.html
.. _git foundation: http://matthew-brett.github.io/pydagogue/foundation.html
.. _deleting main on github: http://matthew-brett.github.io/pydagogue/gh_delete_master.html
.. _rebase without tears: http://matthew-brett.github.io/pydagogue/rebase_without_tears.html
.. _resolving a merge: http://schacon.github.io/git/user-manual.html#resolving-a-merge
.. _ipython git workflow: https://mail.python.org/pipermail/ipython-dev/2010-October/005632.html
.. _ipython notebook on using git in science: https://nbviewer.jupyter.org/github/fperez/reprosw/blob/master/Version%20Control.ipynb

.. other stuff
.. |emdash| unicode:: U+02014
.. vim: ft=rst
.. _`Astropy GitHub`: https://github.com/astropy/astropy
